
# THIS FILE IS GENERATED FROM THEANO SETUP.PY
short_version = '0.4.1rc1'
version = '0.4.1rc1'
hg_revision = 'c00fd7118b57'
full_version = '0.4.1rc1.dev-%(hg_revision)s' % {'hg_revision': hg_revision}
release = False

if not release:
    version = full_version
